from __future__ import annotations

import re
from typing import NamedTuple
from typing import Protocol

ERR_RE = re.compile(
    r'^(?P<filename>[^:]+):'
    r'(?P<lineno>[0-9]+):'
    r'(?:(?P<col_offset>[0-9]+):)?'
    r'(?P<msg>.*)$',
)


class Error(NamedTuple):
    filename: str
    lineno: int
    col_offset: int
    msg: str
    disabled: bool = False

    @property
    def line_idx(self) -> int:
        return self.lineno - 1

    @property
    def pos(self) -> tuple[int, int]:
        return self.lineno, self.col_offset

    def render(self) -> str:
        if self.disabled:
            return f'??:??: {self.msg}'
        else:
            return f'{self.lineno}:{self.col_offset}: {self.msg}'


def parse_generic_output(s: str) -> tuple[Error, ...]:
    ret = []
    for line in s.splitlines():
        match = ERR_RE.match(line)
        if match is not None:
            error = Error(
                filename=match['filename'],
                lineno=int(match['lineno']),
                col_offset=int(match['col_offset'] or '1'),
                msg=match['msg'].strip(),
            )
            ret.append(error)
    return tuple(ret)


class Linter(Protocol):
    def command(self, filename: str, scope: str) -> tuple[str, ...] | None: ...
    def parse(self, filename: str, output: str) -> tuple[Error, ...]: ...



from __future__ import annotations

import curses
from functools import cached_property


def line_x(x: int, width: int) -> int:
    if x + 1 < width:
        return 0
    elif width == 1:
        return x
    else:
        margin = min(width - 3, 6)
        return (
            width - margin - 2 +
            (x + 1 - width) //
            (width - margin - 2) *
            (width - margin - 2)
        )


def scrolled_line(s: str, x: int, width: int) -> str:
    l_x = line_x(x, width)
    if l_x:
        s = f'«{s[l_x + 1:]}'
        if len(s) > width:
            return f'{s[:width - 1]}»'
        else:
            return s.ljust(width)
    elif len(s) > width:
        return f'{s[:width - 1]}»'
    else:
        return s.ljust(width)


class _CalcWidth:
    @cached_property
    def _window(self) -> curses._CursesWindow:
        return curses.newwin(1, 10)

    def wcwidth(self, c: str) -> int:
        self._window.addstr(0, 0, c)
        return self._window.getyx()[1]


wcwidth = _CalcWidth().wcwidth
del _CalcWidth



from __future__ import annotations

import argparse
from typing import Sequence

from babi.highlight import Compiler
from babi.highlight import Grammars
from babi.highlight import highlight_line
from babi.theme import Style
from babi.theme import Theme
from babi.user_data import prefix_data
from babi.user_data import xdg_config


def print_styled(s: str, style: Style) -> None:
    color_s = ''
    undo_s = ''
    if style.fg is not None:
        color_s += '\x1b[38;2;{r};{g};{b}m'.format(**style.fg._asdict())
        undo_s += '\x1b[39m'
    if style.bg is not None:
        color_s += '\x1b[48;2;{r};{g};{b}m'.format(**style.bg._asdict())
        undo_s += '\x1b[49m'
    if style.b:
        color_s += '\x1b[1m'
        undo_s += '\x1b[22m'
    if style.i:
        color_s += '\x1b[3m'
        undo_s += '\x1b[23m'
    if style.u:
        color_s += '\x1b[4m'
        undo_s += '\x1b[24m'
    print(f'{color_s}{s}{undo_s}', end='', flush=True)


def _highlight_output(theme: Theme, compiler: Compiler, filename: str) -> int:
    state = compiler.root_state

    if theme.default.bg is not None:
        print('\x1b[48;2;{r};{g};{b}m'.format(**theme.default.bg._asdict()))
    with open(filename, encoding='UTF-8') as f:
        for line_idx, line in enumerate(f):
            first_line = line_idx == 0
            state, regions = highlight_line(compiler, state, line, first_line)
            for start, end, scope in regions:
                print_styled(line[start:end], theme.select(scope))
    print('\x1b[m', end='')
    return 0


def main(argv: Sequence[str] | None = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('--theme', default=xdg_config('theme.json'))
    parser.add_argument('--grammar-dir', default=prefix_data('grammar_v1'))
    parser.add_argument('filename')
    args = parser.parse_args(argv)

    with open(args.filename, encoding='UTF-8') as f:
        first_line = next(f, '')

    theme = Theme.from_filename(args.theme)

    grammars = Grammars(args.grammar_dir)
    compiler = grammars.compiler_for_file(args.filename, first_line)

    return _highlight_output(theme, compiler, args.filename)


if __name__ == '__main__':
    raise SystemExit(main())



from __future__ import annotations

from babi import linting


class Flake8:
    def command(self, filename: str, scope: str) -> tuple[str, ...] | None:
        if scope != 'source.python':
            return None
        else:
            # TODO: forbid color output
            return ('flake8', filename)

    def parse(self, filename: str, output: str) -> tuple[linting.Error, ...]:
        return tuple(
            error._replace(msg=f'[flake8] {error.msg}')
            for error in linting.parse_generic_output(output)
        )



from __future__ import annotations

import curses
import enum
from typing import TYPE_CHECKING

from babi.horizontal_scrolling import line_x
from babi.horizontal_scrolling import scrolled_line

if TYPE_CHECKING:
    from babi.main import Screen  # XXX: circular

PromptResult = enum.Enum('PromptResult', 'CANCELLED')


class Prompt:
    def __init__(self, screen: Screen, prompt: str, lst: list[str]) -> None:
        self._screen = screen
        self._prompt = prompt
        self._lst = lst
        self._y = len(lst) - 1
        self._x = len(self._s)

    @property
    def _s(self) -> str:
        return self._lst[self._y]

    @_s.setter
    def _s(self, s: str) -> None:
        self._lst[self._y] = s

    def _render_prompt(self, *, base: str | None = None) -> None:
        base = base or self._prompt
        if not base or self._screen.layout.status.width < 7:
            prompt_s = ''
        elif len(base) > self._screen.layout.status.width - 6:
            prompt_s = f'{base[:self._screen.layout.status.width - 7]}…: '
        else:
            prompt_s = f'{base}: '
        width = self._screen.layout.status.width - len(prompt_s)
        line = scrolled_line(self._s, self._x, width)
        cmd = f'{prompt_s}{line}'
        prompt_line = self._screen.layout.status.y
        self._screen.stdscr.insstr(prompt_line, 0, cmd, curses.A_REVERSE)
        x = len(prompt_s) + self._x - line_x(self._x, width)
        self._screen.stdscr.move(prompt_line, x)

    def _up(self) -> None:
        self._y = max(0, self._y - 1)
        self._x = len(self._s)

    def _down(self) -> None:
        self._y = min(len(self._lst) - 1, self._y + 1)
        self._x = len(self._s)

    def _right(self) -> None:
        self._x = min(len(self._s), self._x + 1)

    def _left(self) -> None:
        self._x = max(0, self._x - 1)

    def _home(self) -> None:
        self._x = 0

    def _end(self) -> None:
        self._x = len(self._s)

    def _ctrl_left(self) -> None:
        if self._x <= 1:
            self._x = 0
        else:
            self._x -= 1
            tp = self._s[self._x - 1].isalnum()
            while self._x > 0 and tp == self._s[self._x - 1].isalnum():
                self._x -= 1

    def _ctrl_right(self) -> None:
        if self._x >= len(self._s) - 1:
            self._x = len(self._s)
        else:
            self._x += 1
            tp = self._s[self._x].isalnum()
            while self._x < len(self._s) and tp == self._s[self._x].isalnum():
                self._x += 1

    def _backspace(self) -> None:
        if self._x > 0:
            self._s = self._s[:self._x - 1] + self._s[self._x:]
            self._x -= 1

    def _delete(self) -> None:
        if self._x < len(self._s):
            self._s = self._s[:self._x] + self._s[self._x + 1:]

    def _cut_to_end(self) -> None:
        self._s = self._s[:self._x]

    def _resize(self) -> None:
        self._screen.resize()

    def _check_failed(self, idx: int, s: str) -> tuple[bool, int]:
        failed = False
        for search_idx in range(idx, -1, -1):
            if s in self._lst[search_idx]:
                idx = self._y = search_idx
                self._x = self._lst[search_idx].index(s)
                break
        else:
            failed = True
        return failed, idx

    def _reverse_search(self) -> None | str | PromptResult:
        reverse_s = ''
        idx = self._y
        while True:
            fail, idx = self._check_failed(idx, reverse_s)

            if fail:
                base = f'{self._prompt}(failed reverse-search)`{reverse_s}`'
            else:
                base = f'{self._prompt}(reverse-search)`{reverse_s}`'

            self._render_prompt(base=base)

            key = self._screen.get_char()
            if key.keyname == b'KEY_RESIZE':
                self._screen.resize()
            elif key.keyname == b'KEY_BACKSPACE':
                reverse_s = reverse_s[:-1]
            elif key.keyname == b'^R':
                idx = max(0, idx - 1)
            elif key.keyname == b'^C':
                return self._screen.status.cancelled()
            elif key.keyname == b'^M':
                return self._s
            elif key.keyname == b'STRING':
                assert isinstance(key.wch, str), key.wch
                for c in key.wch:
                    reverse_s += c
                    failed, idx = self._check_failed(idx, reverse_s)
            else:
                self._x = len(self._s)
                return None

    def _cancel(self) -> PromptResult:
        return self._screen.status.cancelled()

    def _submit(self) -> str:
        return self._s

    DISPATCH = {
        # movement
        b'KEY_UP': _up,
        b'KEY_DOWN': _down,
        b'KEY_RIGHT': _right,
        b'KEY_LEFT': _left,
        b'KEY_HOME': _home,
        b'^A': _home,
        b'KEY_END': _end,
        b'^E': _end,
        b'kRIT5': _ctrl_right,
        b'kLFT5': _ctrl_left,
        # editing
        b'KEY_BACKSPACE': _backspace,
        b'KEY_DC': _delete,
        b'^K': _cut_to_end,
        # misc
        b'KEY_RESIZE': _resize,
        b'^R': _reverse_search,
        b'^M': _submit,
        b'^C': _cancel,
    }

    def _c(self, c: str) -> None:
        self._s = self._s[:self._x] + c + self._s[self._x:]
        self._x += len(c)

    def run(self) -> PromptResult | str:
        while True:
            self._render_prompt()

            key = self._screen.get_char()
            if key.keyname in Prompt.DISPATCH:
                ret = Prompt.DISPATCH[key.keyname](self)
                if ret is not None:
                    return ret
            elif key.keyname == b'STRING':
                assert isinstance(key.wch, str), key.wch
                self._c(key.wch)



from __future__ import annotations

from typing import NamedTuple
from typing import Protocol
from typing import Tuple

from babi.buf import Buf


class HL(NamedTuple):
    x: int
    end: int
    attr: int


HLs = Tuple[HL, ...]


class RegionsMapping(Protocol):
    def __getitem__(self, idx: int) -> HLs: ...


class FileHL(Protocol):
    @property
    def include_edge(self) -> bool: ...
    @property
    def regions(self) -> RegionsMapping: ...
    def highlight_until(self, lines: Buf, idx: int) -> None: ...
    def register_callbacks(self, buf: Buf) -> None: ...



from __future__ import annotations

import contextlib
import curses
import enum
import hashlib
import importlib.metadata
import os
import re
import signal
import sre_parse
import subprocess
import sys
from types import FrameType
from typing import Callable
from typing import Generator
from typing import NamedTuple
from typing import Pattern

from babi import linting
from babi.color_manager import ColorManager
from babi.dim import Dim
from babi.file import Action
from babi.file import File
from babi.file import get_lines
from babi.file import OPEN_SETTINGS
from babi.history import History
from babi.hl.syntax import Syntax
from babi.linters.flake8 import Flake8
from babi.linters.pre_commit import PreCommit
from babi.perf import Perf
from babi.proc import graceful_terminate
from babi.prompt import Prompt
from babi.prompt import PromptResult
from babi.status import Status

VERSION_STR = f'babi v{importlib.metadata.version("babi")}'

# TODO: find a place to populate these, surely there's a database somewhere
SEQUENCE_KEYNAME = {
    '\x1bOH': b'KEY_HOME',
    '\x1bOF': b'KEY_END',
    '\x1b[1~': b'KEY_HOME',
    '\x1b[4~': b'KEY_END',
    '\x1b[1;2A': b'KEY_SR',
    '\x1b[1;2B': b'KEY_SF',
    '\x1b[1;2C': b'KEY_SRIGHT',
    '\x1b[1;2D': b'KEY_SLEFT',
    '\x1b[1;2H': b'KEY_SHOME',
    '\x1b[1;2F': b'KEY_SEND',
    '\x1b[5;2~': b'KEY_SPREVIOUS',
    '\x1b[6;2~': b'KEY_SNEXT',
    '\x1b[1;3A': b'kUP3',  # M-Up
    '\x1b[1;3B': b'kDN3',  # M-Down
    '\x1b[1;3C': b'kRIT3',  # M-Right
    '\x1b[1;3D': b'kLFT3',  # M-Left
    '\x1b[1;5A': b'kUP5',  # ^Up
    '\x1b[1;5B': b'kDN5',  # ^Down
    '\x1b[1;5C': b'kRIT5',  # ^Right
    '\x1b[1;5D': b'kLFT5',  # ^Left
    '\x1b[1;5H': b'kHOM5',  # ^Home
    '\x1b[1;5F': b'kEND5',  # ^End
    '\x1b[1;6A': b'kUP6',  # Shift + ^Up
    '\x1b[1;6B': b'kDN6',  # Shift + ^Down
    '\x1b[1;6C': b'kRIT6',  # Shift + ^Right
    '\x1b[1;6D': b'kLFT6',  # Shift + ^Left
    '\x1b[1;6H': b'kHOM6',  # Shift + ^Home
    '\x1b[1;6F': b'kEND6',  # Shift + ^End
    '\x1b[~': b'KEY_BTAB',  # Shift + Tab
}
KEYNAME_REWRITE = {
    # windows-curses: numeric pad arrow keys
    # - some overlay keyboards pick these as well
    # - in xterm it seems these are mapped automatically
    b'KEY_A2': b'KEY_UP',
    b'KEY_C2': b'KEY_DOWN',
    b'KEY_B3': b'KEY_RIGHT',
    b'KEY_B1': b'KEY_LEFT',
    b'PADSTOP': b'KEY_DC',
    b'KEY_A3': b'KEY_PPAGE',
    b'KEY_C3': b'KEY_NPAGE',
    b'KEY_A1': b'KEY_HOME',
    b'KEY_C1': b'KEY_END',
    # windows-curses: map to our M- names
    b'ALT_U': b'M-u',
    # windows-curses: arguably these names are better than the xterm names
    b'CTL_UP': b'kUP5',
    b'CTL_DOWN': b'kDN5',
    b'CTL_RIGHT': b'kRIT5',
    b'CTL_LEFT': b'kLFT5',
    b'CTL_HOME': b'kHOM5',
    b'CTL_END': b'kEND5',
    b'ALT_RIGHT': b'kRIT3',
    b'ALT_LEFT': b'kLFT3',
    b'ALT_E': b'M-e',
    # windows-curses: idk why these are different
    b'KEY_SUP': b'KEY_SR',
    b'KEY_SDOWN': b'KEY_SF',
    # macos: (sends this for backspace key, others interpret this as well)
    b'^?': b'KEY_BACKSPACE',
    # linux, perhaps others
    b'^H': b'KEY_BACKSPACE',  # ^Backspace on my keyboard
    b'^D': b'KEY_DC',
    b'PADENTER': b'^M',  # Enter on numpad
    b'KEY_ENTER': b'^M',  # Enter, but in `fn` mode
}

LINTER_TYPES: tuple[type[linting.Linter], ...] = (PreCommit, Flake8)


def _get_wch_with_retry(stdscr: curses._CursesWindow) -> str | int:
    while True:
        try:
            return stdscr.get_wch()
        except curses.error:  # pragma: no cover (error during signals?)
            pass


class EditResult(enum.Enum):
    EXIT = enum.auto()
    EXIT_ALL = enum.auto()
    EXIT_ALL_FORCE = enum.auto()
    NEXT = enum.auto()
    PREV = enum.auto()
    OPEN = enum.auto()


class Command(NamedTuple):
    callback: Callable[[Screen, list[str]], EditResult | None]
    nargs: str | int = 0


class Key(NamedTuple):
    wch: int | str
    keyname: bytes


class FileInfo(NamedTuple):
    filename: str | None
    initial_line: int
    is_stdin: bool


class Layout(NamedTuple):
    header: Dim
    file: Dim
    status: Dim
    lint_errors: Dim


class Screen:
    def __init__(
            self,
            stdscr: curses._CursesWindow,
            file_infos: list[FileInfo],
            perf: Perf,
    ) -> None:
        self.stdscr = stdscr
        self.syntax = Syntax.from_screen(stdscr, ColorManager.make())
        self.files = [
            File(
                info.filename,
                info.initial_line,
                self.syntax,
                is_stdin=info.is_stdin,
            )
            for info in file_infos
        ]
        self.i = 0
        self.history = History()
        self.perf = perf
        self.layout = self._layout_from_current_screen()
        self.status = Status()
        self.cut_buffer: tuple[str, ...] = ()
        self.cut_selection = False
        self._buffered_input: int | str | None = None
        self._retheme = False
        self._linters = tuple(tp() for tp in LINTER_TYPES)

    @property
    def file(self) -> File:
        return self.files[self.i]

    def _draw_header(self, dim: Dim) -> None:
        filename = self.file.filename or '<<new file>>'
        if self.file.modified:
            filename += ' *'
        if len(self.files) > 1:
            files = f'[{self.i + 1}/{len(self.files)}] '
            version_width = len(VERSION_STR) + 2 + len(files)
        else:
            files = ''
            version_width = len(VERSION_STR) + 2
        centered = filename.center(dim.width)[version_width:]
        s = f' {VERSION_STR} {files}{centered}'
        self.stdscr.insstr(0, 0, s, curses.A_REVERSE)

    def _get_sequence_home_end(self, wch: str) -> str:
        try:
            c = self.stdscr.get_wch()
        except curses.error:
            return wch
        else:
            if isinstance(c, int) or c not in 'HF':
                self._buffered_input = c
                return wch
            else:
                return f'{wch}{c}'

    def _get_sequence_bracketed(self, wch: str) -> str:
        for _ in range(3):  # [0-9]{1,2};
            try:
                c = self.stdscr.get_wch()
            except curses.error:
                return wch
            else:
                if isinstance(c, int):
                    self._buffered_input = c
                    return wch
                else:
                    wch += c
                    if c == ';':
                        break
        else:
            return wch  # unexpected input while searching for `;`

        for _ in range(2):  # [0-9].
            try:
                c = self.stdscr.get_wch()
            except curses.error:
                return wch
            else:
                if isinstance(c, int):
                    self._buffered_input = c
                    return wch
                else:
                    wch += c

        return wch

    def _get_sequence(self, wch: str) -> str:
        self.stdscr.nodelay(True)
        try:
            c = self.stdscr.get_wch()
        except curses.error:
            return wch
        else:
            if isinstance(c, int):  # M-BSpace
                return f'{wch}({c})'  # TODO
            elif c == 'O':
                return self._get_sequence_home_end(f'{wch}O')
            elif c == '[':
                return self._get_sequence_bracketed(f'{wch}[')
            else:
                return f'{wch}{c}'
        finally:
            self.stdscr.nodelay(False)

    def _get_string(self, wch: str) -> str:
        self.stdscr.nodelay(True)
        try:
            while True:
                try:
                    c = self.stdscr.get_wch()
                    if isinstance(c, str) and c.isprintable():
                        wch += c
                    else:
                        self._buffered_input = c
                        break
                except curses.error:
                    break
        finally:
            self.stdscr.nodelay(False)
        return wch

    def _get_char(self) -> Key:
        if self._buffered_input is not None:
            wch, self._buffered_input = self._buffered_input, None
        elif self._retheme:
            self._retheme = False
            return Key(-1, b'RETHEME')
        else:
            wch = _get_wch_with_retry(self.stdscr)
        if isinstance(wch, str) and wch == '\x1b':
            wch = self._get_sequence(wch)
            if len(wch) == 2:
                return Key(wch, f'M-{wch[1]}'.encode())
            elif len(wch) > 1:
                keyname = SEQUENCE_KEYNAME.get(wch, b'unknown')
                return Key(wch, keyname)
        elif isinstance(wch, str) and wch.isprintable():
            wch = self._get_string(wch)
            return Key(wch, b'STRING')

        key = wch if isinstance(wch, int) else ord(wch)
        keyname = curses.keyname(key)
        keyname = KEYNAME_REWRITE.get(keyname, keyname)
        return Key(wch, keyname)

    def get_char(self) -> Key:
        self.perf.end()
        ret = self._get_char()
        self.perf.start(ret.keyname.decode())
        return ret

    def draw(self) -> None:
        self._draw_header(self.layout.header)
        self.file.draw(self.stdscr, self.layout.file)
        self.status.draw(self.stdscr, self.layout.status)
        self.file.lint_errors.draw(self.stdscr, self.layout.lint_errors)

    def _layout_from_current_screen(self) -> Layout:
        if curses.LINES >= 10:
            lint_errors_height = min(len(self.file.lint_errors.errors), 3)
        else:
            lint_errors_height = 0

        file_y = 0
        file_height = curses.LINES - lint_errors_height

        if curses.LINES > 2:
            file_y += 1
            file_height -= 2
        elif curses.LINES > 1:
            file_height -= 1

        status_y = curses.LINES - lint_errors_height - 1

        return Layout(
            header=Dim(x=0, y=0, width=curses.COLS, height=1),
            file=Dim(x=0, y=file_y, width=curses.COLS, height=file_height),
            status=Dim(x=0, y=status_y, width=curses.COLS, height=1),
            lint_errors=Dim(
                x=0,
                y=curses.LINES - lint_errors_height,
                width=curses.COLS,
                height=lint_errors_height,
            ),
        )

    def resize(self) -> None:
        curses.update_lines_cols()
        self.layout = self._layout_from_current_screen()
        self.file.buf.scroll_screen_if_needed(self.layout.file)
        self.draw()

    def quick_prompt(
            self,
            prompt: str,
            opt_strs: tuple[str, ...],
    ) -> str | PromptResult:
        opts = {opt[0] for opt in opt_strs}
        while True:
            x = 0
            prompt_line = self.layout.status.y

            def _write(s: str, *, attr: int = curses.A_REVERSE) -> None:
                nonlocal x

                if x >= self.layout.status.width:
                    return
                self.stdscr.insstr(prompt_line, x, s, attr)
                x += len(s)

            _write(prompt)
            _write(' [')
            for i, opt_str in enumerate(opt_strs):
                _write(opt_str[0], attr=curses.A_REVERSE | curses.A_BOLD)
                _write(opt_str[1:])
                if i != len(opt_strs) - 1:
                    _write(', ')
            _write(']?')

            if x < self.layout.status.width - 1:
                s = ' ' * (self.layout.status.width - x)
                self.stdscr.insstr(prompt_line, x, s, curses.A_REVERSE)
                x += 1
            else:
                x = self.layout.status.width - 1
                self.stdscr.insstr(prompt_line, x, '…', curses.A_REVERSE)

            self.stdscr.move(prompt_line, x)

            key = self.get_char()
            if key.keyname == b'KEY_RESIZE':
                self.resize()
            elif key.keyname == b'^C':
                return self.status.cancelled()
            elif isinstance(key.wch, str) and key.wch.lower() in opts:
                return key.wch.lower()

    def prompt(
            self,
            prompt: str,
            *,
            allow_empty: bool = False,
            history: str | None = None,
            default_prev: bool = False,
            default: str | None = None,
    ) -> str | PromptResult:
        default = default or ''
        self.status.clear()
        if history is not None:
            history_data = [*self.history.data[history], default]
            if default_prev and history in self.history.prev:
                prompt = f'{prompt} [{self.history.prev[history]}]'
        else:
            history_data = [default]

        ret = Prompt(self, prompt, history_data).run()

        if ret is not PromptResult.CANCELLED and history is not None:
            if ret:  # only put non-empty things in history
                history_lst = self.history.data[history]
                if not history_lst or history_lst[-1] != ret:
                    history_lst.append(ret)
                self.history.prev[history] = ret
            elif default_prev and history in self.history.prev:
                return self.history.prev[history]

        if not allow_empty and not ret:
            return self.status.cancelled()
        else:
            return ret

    def go_to_line(self) -> None:
        response = self.prompt('enter line number')
        if response is not PromptResult.CANCELLED:
            try:
                lineno = int(response)
            except ValueError:
                self.status.update(f'not an integer: {response!r}')
            else:
                self.file.go_to_line(lineno, self.layout.file)

    def current_position(self) -> None:
        line = f'line {self.file.buf.y + 1}'
        col = f'col {self.file.buf.x + 1}'
        line_count = max(len(self.file.buf) - 1, 1)
        lines_word = 'line' if line_count == 1 else 'lines'
        self.status.update(f'{line}, {col} (of {line_count} {lines_word})')

    def cut(self) -> None:
        if self.file.selection.start:
            self.cut_buffer = self.file.cut_selection(self.layout.file)
            self.cut_selection = True
        else:
            self.cut_buffer = self.file.cut(self.cut_buffer)
            self.cut_selection = False

    def uncut(self) -> None:
        if self.cut_selection:
            self.file.uncut_selection(self.cut_buffer, self.layout.file)
        else:
            self.file.uncut(self.cut_buffer, self.layout.file)

    def _get_search_re(self, prompt: str) -> Pattern[str] | PromptResult:
        response = self.prompt(prompt, history='search', default_prev=True)
        if response is PromptResult.CANCELLED:
            return response
        try:
            return re.compile(response)
        except re.error:
            self.status.update(f'invalid regex: {response!r}')
            return PromptResult.CANCELLED

    def _undo_redo(
            self,
            op: str,
            from_stack: list[Action],
            to_stack: list[Action],
    ) -> None:
        if not from_stack:
            self.status.update(f'nothing to {op}!')
        else:
            action = from_stack.pop()
            to_stack.append(action.apply(self.file))
            self.file.buf.scroll_screen_if_needed(self.layout.file)
            self.status.update(f'{op}: {action.name}')
            self.file.selection.clear()

    def undo(self) -> None:
        self._undo_redo('undo', self.file.undo_stack, self.file.redo_stack)

    def redo(self) -> None:
        self._undo_redo('redo', self.file.redo_stack, self.file.undo_stack)

    def search(self) -> None:
        response = self._get_search_re('search')
        if response is not PromptResult.CANCELLED:
            self.file.search(response, self.status, self.layout.file)

    def replace(self) -> None:
        search_response = self._get_search_re('search (to replace)')
        if search_response is not PromptResult.CANCELLED:
            response = self.prompt(
                'replace with', history='replace', allow_empty=True,
            )
            if response is not PromptResult.CANCELLED:
                try:
                    sre_parse.parse_template(response, search_response)
                except re.error:
                    self.status.update('invalid replacement string')
                else:
                    self.file.replace(self, search_response, response)

    def lint(self) -> None:
        if sys.platform == 'win32':  # pragma: win32 cover
            self.status.update('this feature is not available on windows')
            return

        if self.file.filename is None:
            self.status.update('file has not been saved yet!')
            return

        if self.file.modified:
            response = self.quick_prompt(
                'file must be saved before linting - save', ('yes', 'no'),
            )
            if response != 'y':
                self.status.cancelled()
                return
            elif self.save() is PromptResult.CANCELLED:
                return

        self.status.update('linting...')
        self.draw()
        self.file.move_cursor(self.stdscr, self.layout.file)
        self.stdscr.refresh()

        for linter in self._linters:
            cmd = linter.command(self.file.filename, self.file.root_scope)

            # linter does not support this file / scope
            if cmd is None:
                continue

            try:
                proc = subprocess.Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.DEVNULL,
                    text=True,
                )
            except OSError:  # command does not exist or isn't runnable
                continue
            else:
                break
        else:
            self.status.update('no linters available!')
            return

        try:
            sigquit_orig = signal.signal(signal.SIGQUIT, signal.SIG_IGN)
            curses.cbreak()
            out, _ = proc.communicate()
        except KeyboardInterrupt:
            graceful_terminate(proc)
            self.status.cancelled()
            return
        finally:
            curses.raw()
            signal.signal(signal.SIGQUIT, sigquit_orig)

        errors = linter.parse(self.file.filename, out)
        errors = tuple(sorted(errors))

        # reload the file if it changed
        before = self.file.sha256
        self.file.reload(self.status, self.layout.file)
        after = self.file.sha256

        if after != before:
            formatted = ' (and formatted)'
        else:
            formatted = ''

        self.file.lint_errors.set_errors(errors)
        if not errors:
            self.status.update(f'linted!{formatted}')
        else:
            self.status.update(f'{len(errors)} error(s){formatted}')

        self.layout = self._layout_from_current_screen()

    def lint_focus(self) -> None:
        self.file.lint_errors.focus(self)

    def lint_previous_error(self) -> None:
        self.file.lint_errors.previous_error(self)

    def lint_next_error(self) -> None:
        self.file.lint_errors.next_error(self)

    def _command_w(self, args: list[str]) -> None:
        self.save()

    def _command_wq(self, args: list[str]) -> EditResult:
        self.save()
        return EditResult.EXIT

    def _command_sort(self, args: list[str]) -> None:
        if self.file.selection.start:
            self.file.sort_selection(self.layout.file)
        else:
            self.file.sort(self.layout.file)
        self.status.update('sorted!')

    def _command_sort_bang(self, args: list[str]) -> None:
        if self.file.selection.start:
            self.file.sort_selection(self.layout.file, reverse=True)
        else:
            self.file.sort(self.layout.file, reverse=True)
        self.status.update('sorted!')

    def _command_tabsize(self, args: list[str]) -> None:
        tab_size, = args
        try:
            parsed_tab_size = int(tab_size)
        except ValueError:
            self.status.update(f'invalid size: {tab_size}')
        else:
            if parsed_tab_size <= 0:
                self.status.update(f'invalid size: {parsed_tab_size}')
            else:
                for file in self.files:
                    file.buf.set_tab_size(parsed_tab_size)
                self.status.update('updated!')

    def _command_expandtabs(self, args: list[str]) -> None:
        for file in self.files:
            file.buf.expandtabs = True
        self.status.update('updated!')

    def _command_noexpandtabs(self, args: list[str]) -> None:
        for file in self.files:
            file.buf.expandtabs = False
        self.status.update('updated!')

    def _command_comment(self, args: list[str]) -> None:
        if len(args) == 1:
            comment, = args
        else:
            comment = '#'
        if self.file.selection.start:
            self.file.toggle_comment_selection(comment)
        else:
            self.file.toggle_comment(comment)

    def _command_reload(self, args: list[str]) -> None:
        if self.file.filename is None:
            self.status.update('file has not been saved yet!')
            return

        if self.file.modified:
            response = self.quick_prompt(
                'reload will discard changes - continue', ('yes', 'no'),
            )
            if response != 'y':
                return

        self.file.reload(self.status, self.layout.file)

    def _command_retheme(self, args: list[str]) -> None:
        self.syntax = Syntax.from_screen(self.stdscr, ColorManager.make())
        for file in self.files:
            file.reload_theme(self.syntax)

    COMMANDS = {
        ':qall': Command(lambda self, args: EditResult.EXIT_ALL),
        ':qall!': Command(lambda self, args: EditResult.EXIT_ALL_FORCE),
        ':q': Command(lambda self, args: self.quit_save_modified()),
        ':q!': Command(lambda self, args: EditResult.EXIT),
        ':w': Command(_command_w),
        ':wq': Command(_command_wq),
        ':sort': Command(_command_sort),
        ':sort!': Command(_command_sort_bang),
        ':tabstop': Command(_command_tabsize, nargs=1),
        ':tabsize': Command(_command_tabsize, nargs=1),
        ':expandtabs': Command(_command_expandtabs),
        ':noexpandtabs': Command(_command_noexpandtabs),
        ':comment': Command(_command_comment, nargs='?'),
        ':reload': Command(_command_reload),
        ':retheme': Command(_command_retheme),
    }

    def command(self) -> EditResult | None:
        response = self.prompt('', history='command')
        if response is PromptResult.CANCELLED:
            return None

        cmd, *args = response.split()
        if cmd not in self.COMMANDS:
            self.status.update(f'invalid command: {response}')
            return None

        command = self.COMMANDS[cmd]
        if isinstance(command.nargs, int):
            if len(args) != command.nargs:
                self.status.update(
                    f'`{cmd}`: expected {command.nargs} args '
                    f'but got {len(args)}',
                )
                return None
        elif command.nargs == '?':
            if len(args) not in {0, 1}:
                self.status.update(
                    f'`{cmd}`: expected 0 or 1 args but got {len(args)}',
                )
                return None
        else:
            raise NotImplementedError(command)

        return command.callback(self, args)

    def save(self) -> PromptResult | None:
        self.file.finalize_previous_action()

        # TODO: maybe use mtime / stat as a shortcut for hashing below
        # TODO: strip trailing whitespace?
        # TODO: save atomically?
        if self.file.filename is None:
            filename = self.prompt('enter filename')
            if filename is PromptResult.CANCELLED:
                return PromptResult.CANCELLED
            else:
                self.file.filename = filename

        if not os.path.isfile(self.file.filename):
            sha256: str | None = None
        else:
            try:
                with open(self.file.filename, **OPEN_SETTINGS) as f:
                    *_, sha256 = get_lines(f)
            except UnicodeDecodeError:
                # instead of crashing, show "changed on disk" error
                sha256 = 'error'

        contents = self.file.nl.join(self.file.buf)
        sha256_to_save = hashlib.sha256(contents.encode()).hexdigest()

        # the file on disk is the same as when we opened it
        if sha256 not in (None, self.file.sha256, sha256_to_save):
            self.status.update('(file changed on disk, not implemented)')
            return PromptResult.CANCELLED

        try:
            dir_path = os.path.dirname(os.path.abspath(self.file.filename))
            os.makedirs(dir_path, exist_ok=True)
            with open(self.file.filename, 'w', **OPEN_SETTINGS) as f:
                f.write(contents)
        except OSError as e:
            self.status.update(f'cannot save file: {e}')
            return PromptResult.CANCELLED

        self.file.modified = False
        self.file.sha256 = sha256_to_save
        num_lines = len(self.file.buf) - 1
        lines = 'lines' if num_lines != 1 else 'line'
        self.status.update(f'saved! ({num_lines} {lines} written)')
        self.file.reset_modified_state()
        return None

    def save_filename(self) -> PromptResult | None:
        response = self.prompt('enter filename', default=self.file.filename)
        if response is PromptResult.CANCELLED:
            return PromptResult.CANCELLED
        else:
            self.file.filename = response
            return self.save()

    def open_file(self) -> EditResult | None:
        response = self.prompt('enter filename', history='open')
        if response is not PromptResult.CANCELLED:
            opened = File(
                response,
                0,
                self.syntax,
                is_stdin=False,
            )
            self.files.append(opened)
            return EditResult.OPEN
        else:
            return None

    def quit_save_modified(self) -> EditResult | None:
        if self.file.modified:
            response = self.quick_prompt(
                'file is modified - save', ('yes', 'no'),
            )
            if response == 'y':
                if self.save_filename() is not PromptResult.CANCELLED:
                    return EditResult.EXIT
                else:
                    return None
            elif response == 'n':
                return EditResult.EXIT
            else:
                assert response is PromptResult.CANCELLED
                return None
        return EditResult.EXIT

    def background(self) -> None:
        if sys.platform == 'win32':  # pragma: win32 cover
            self.status.update('cannot run babi in background on Windows')
        else:  # pragma: win32 no cover
            curses.endwin()
            os.kill(os.getpid(), signal.SIGSTOP)
            self.stdscr = _init_screen()
            self.resize()

    def retheme(self) -> None:
        self._command_retheme([])

    DISPATCH = {
        b'RETHEME': retheme,
        b'KEY_RESIZE': resize,
        b'^_': go_to_line,
        b'^C': current_position,
        b'^K': cut,
        b'^U': uncut,
        b'M-u': undo,
        b'M-U': redo,
        b'M-e': redo,
        b'^W': search,
        b'^\\': replace,
        b'^T': lint,
        b'M-t': lint_focus,
        b'kUP6': lint_previous_error,
        b'kDN6': lint_next_error,
        b'^[': command,
        b'^S': save,
        b'^O': save_filename,
        b'^X': quit_save_modified,
        b'^P': open_file,
        b'kLFT3': lambda screen: EditResult.PREV,
        b'kRIT3': lambda screen: EditResult.NEXT,
        b'^Z': background,
    }

    @contextlib.contextmanager
    def retheme_handler(self) -> Generator[None]:
        if sys.platform == 'win32':  # pragma: win32 cover
            yield  # no signal handling on windows!
        else:
            def sigusr1_handler(signum: int, frame: FrameType | None) -> None:
                self._retheme = True
                os.kill(os.getpid(), signal.SIGWINCH)

            orig = signal.signal(signal.SIGUSR1, sigusr1_handler)
            try:
                yield
            finally:
                signal.signal(signal.SIGUSR1, orig)


def _init_screen() -> curses._CursesWindow:
    # set the escape delay so curses does not pause waiting for sequences
    if (
            sys.version_info >= (3, 9) and
            hasattr(curses, 'set_escdelay')
    ):  # pragma: >=3.9 cover
        curses.set_escdelay(25)
    else:  # pragma: <3.9 cover
        os.environ.setdefault('ESCDELAY', '25')

    stdscr = curses.initscr()
    curses.noecho()
    # <enter> is not transformed into '\n' so it can be differentiated from ^J
    curses.nonl()
    # ^S / ^Q / ^Z / ^\ are passed through
    curses.raw()
    stdscr.keypad(True)

    with contextlib.suppress(curses.error):
        curses.start_color()
        curses.use_default_colors()
    return stdscr


@contextlib.contextmanager
def make_stdscr() -> Generator[curses._CursesWindow]:
    """essentially `curses.wrapper` but split out to implement ^Z"""
    try:
        yield _init_screen()
    finally:
        curses.endwin()



from __future__ import annotations

DEFAULT_THEME = {
    '$schema': 'vscode://schemas/color-theme',
    'name': 'Dark (Visual Studio)',
    'colors': {
        'editor.background': '#1E1E1E',
        'editor.foreground': '#D4D4D4',
        'editor.inactiveSelectionBackground': '#3A3D41',
        'editorIndentGuide.background': '#404040',
        'editorIndentGuide.activeBackground': '#707070',
        'editor.selectionHighlightBackground': '#ADD6FF26',
        'list.dropBackground': '#383B3D',
        'activityBarBadge.background': '#007ACC',
        'sideBarTitle.foreground': '#BBBBBB',
        'input.placeholderForeground': '#A6A6A6',
        'settings.textInputBackground': '#292929',
        'settings.numberInputBackground': '#292929',
        'menu.background': '#252526',
        'menu.foreground': '#CCCCCC',
        'statusBarItem.remoteForeground': '#FFF',
        'statusBarItem.remoteBackground': '#16825D',
    },
    'tokenColors': [
        {
            'scope': [
                'meta.embedded',
                'source.groovy.embedded',
            ],
            'settings': {
                'foreground': '#D4D4D4',
            },
        },
        {
            'scope': 'emphasis',
            'settings': {
                'fontStyle': 'italic',
            },
        },
        {
            'scope': 'strong',
            'settings': {
                'fontStyle': 'bold',
            },
        },
        {
            'scope': 'header',
            'settings': {
                'foreground': '#000080',
            },
        },
        {
            'scope': 'comment',
            'settings': {
                'foreground': '#6A9955',
            },
        },
        {
            'scope': 'constant.language',
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': [
                'constant.numeric',
                'entity.name.operator.custom-literal.number',
                'variable.other.enummember',
                'keyword.operator.plus.exponent',
                'keyword.operator.minus.exponent',
            ],
            'settings': {
                'foreground': '#b5cea8',
            },
        },
        {
            'scope': 'constant.regexp',
            'settings': {
                'foreground': '#646695',
            },
        },
        {
            'scope': 'entity.name.tag',
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'entity.name.tag.css',
            'settings': {
                'foreground': '#d7ba7d',
            },
        },
        {
            'scope': 'entity.other.attribute-name',
            'settings': {
                'foreground': '#9cdcfe',
            },
        },
        {
            'scope': [
                'entity.other.attribute-name.class.css',
                'entity.other.attribute-name.class.mixin.css',
                'entity.other.attribute-name.id.css',
                'entity.other.attribute-name.parent-selector.css',
                'entity.other.attribute-name.pseudo-class.css',
                'entity.other.attribute-name.pseudo-element.css',
                'source.css.less entity.other.attribute-name.id',
                'entity.other.attribute-name.attribute.scss',
                'entity.other.attribute-name.scss',
            ],
            'settings': {
                'foreground': '#d7ba7d',
            },
        },
        {
            'scope': 'invalid',
            'settings': {
                'foreground': '#f44747',
            },
        },
        {
            'scope': 'markup.underline',
            'settings': {
                'fontStyle': 'underline',
            },
        },
        {
            'scope': 'markup.bold',
            'settings': {
                'fontStyle': 'bold',
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'markup.heading',
            'settings': {
                'fontStyle': 'bold',
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'markup.italic',
            'settings': {
                'fontStyle': 'italic',
            },
        },
        {
            'scope': 'markup.inserted',
            'settings': {
                'foreground': '#b5cea8',
            },
        },
        {
            'scope': 'markup.deleted',
            'settings': {
                'foreground': '#ce9178',
            },
        },
        {
            'scope': 'markup.changed',
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'punctuation.definition.quote.begin.markdown',
            'settings': {
                'foreground': '#6A9955',
            },
        },
        {
            'scope': 'punctuation.definition.list.begin.markdown',
            'settings': {
                'foreground': '#6796e6',
            },
        },
        {
            'scope': 'markup.inline.raw',
            'settings': {
                'foreground': '#ce9178',
            },
        },
        {
            'name': 'brackets of XML/HTML tags',
            'scope': 'punctuation.definition.tag',
            'settings': {
                'foreground': '#808080',
            },
        },
        {
            'scope': [
                'meta.preprocessor',
                'entity.name.function.preprocessor',
            ],
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'meta.preprocessor.string',
            'settings': {
                'foreground': '#ce9178',
            },
        },
        {
            'scope': 'meta.preprocessor.numeric',
            'settings': {
                'foreground': '#b5cea8',
            },
        },
        {
            'scope': 'meta.structure.dictionary.key.python',
            'settings': {
                'foreground': '#9cdcfe',
            },
        },
        {
            'scope': 'meta.diff.header',
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'storage',
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'storage.type',
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': [
                'storage.modifier',
                'keyword.operator.noexcept',
            ],
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': [
                'string',
                'entity.name.operator.custom-literal.string',
                'meta.embedded.assembly',
            ],
            'settings': {
                'foreground': '#ce9178',
            },
        },
        {
            'scope': 'string.tag',
            'settings': {
                'foreground': '#ce9178',
            },
        },
        {
            'scope': 'string.value',
            'settings': {
                'foreground': '#ce9178',
            },
        },
        {
            'scope': 'string.regexp',
            'settings': {
                'foreground': '#d16969',
            },
        },
        {
            'name': 'String interpolation',
            'scope': [
                'punctuation.definition.template-expression.begin',
                'punctuation.definition.template-expression.end',
                'punctuation.section.embedded',
            ],
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'name': 'Reset JavaScript string interpolation expression',
            'scope': [
                'meta.template.expression',
            ],
            'settings': {
                'foreground': '#d4d4d4',
            },
        },
        {
            'scope': [
                'support.type.vendored.property-name',
                'support.type.property-name',
                'variable.css',
                'variable.scss',
                'variable.other.less',
                'source.coffee.embedded',
            ],
            'settings': {
                'foreground': '#9cdcfe',
            },
        },
        {
            'scope': 'keyword',
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'keyword.control',
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'keyword.operator',
            'settings': {
                'foreground': '#d4d4d4',
            },
        },
        {
            'scope': [
                'keyword.operator.new',
                'keyword.operator.expression',
                'keyword.operator.cast',
                'keyword.operator.sizeof',
                'keyword.operator.alignof',
                'keyword.operator.typeid',
                'keyword.operator.alignas',
                'keyword.operator.instanceof',
                'keyword.operator.logical.python',
                'keyword.operator.wordlike',
            ],
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'keyword.other.unit',
            'settings': {
                'foreground': '#b5cea8',
            },
        },
        {
            'scope': [
                'punctuation.section.embedded.begin.php',
                'punctuation.section.embedded.end.php',
            ],
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'support.function.git-rebase',
            'settings': {
                'foreground': '#9cdcfe',
            },
        },
        {
            'scope': 'constant.sha.git-rebase',
            'settings': {
                'foreground': '#b5cea8',
            },
        },
        {
            'name': 'coloring of the Java import and package identifiers',
            'scope': [
                'storage.modifier.import.java',
                'variable.language.wildcard.java',
                'storage.modifier.package.java',
            ],
            'settings': {
                'foreground': '#d4d4d4',
            },
        },
        {
            'name': 'this.self',
            'scope': 'variable.language',
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'name': 'Function declarations',
            'scope': [
                'entity.name.function',
                'support.function',
                'support.constant.handlebars',
                'source.powershell variable.other.member',
                'entity.name.operator.custom-literal',
            ],
            'settings': {
                'foreground': '#DCDCAA',
            },
        },
        {
            'name': 'Types declaration and references',
            'scope': [
                'meta.return-type',
                'support.class',
                'support.type',
                'entity.name.type',
                'entity.name.namespace',
                'entity.other.attribute',
                'entity.name.scope-resolution',
                'entity.name.class',
                'storage.type.numeric.go',
                'storage.type.byte.go',
                'storage.type.boolean.go',
                'storage.type.string.go',
                'storage.type.uintptr.go',
                'storage.type.error.go',
                'storage.type.rune.go',
                'storage.type.cs',
                'storage.type.generic.cs',
                'storage.type.modifier.cs',
                'storage.type.variable.cs',
                'storage.type.annotation.java',
                'storage.type.generic.java',
                'storage.type.java',
                'storage.type.object.array.java',
                'storage.type.primitive.array.java',
                'storage.type.primitive.java',
                'storage.type.token.java',
                'storage.type.groovy',
                'storage.type.annotation.groovy',
                'storage.type.parameters.groovy',
                'storage.type.generic.groovy',
                'storage.type.object.array.groovy',
                'storage.type.primitive.array.groovy',
                'storage.type.primitive.groovy',
            ],
            'settings': {
                'foreground': '#4EC9B0',
            },
        },
        {
            'name': 'Types declaration and references, TS grammar specific',
            'scope': [
                'meta.type.cast.expr',
                'meta.type.new.expr',
                'support.constant.math',
                'support.constant.dom',
                'support.constant.json',
                'entity.other.inherited-class',
            ],
            'settings': {
                'foreground': '#4EC9B0',
            },
        },
        {
            'name': 'Control flow / Special keywords',
            'scope': [
                'keyword.control',
                'source.cpp keyword.operator.new',
                'keyword.operator.delete',
                'keyword.other.using',
                'keyword.other.operator',
                'entity.name.operator',
            ],
            'settings': {
                'foreground': '#C586C0',
            },
        },
        {
            'name': 'Variable and parameter name',
            'scope': [
                'variable',
                'meta.definition.variable.name',
                'support.variable',
                'entity.name.variable',
            ],
            'settings': {
                'foreground': '#9CDCFE',
            },
        },
        {
            'name': 'Object keys, TS grammar specific',
            'scope': [
                'meta.object-literal.key',
            ],
            'settings': {
                'foreground': '#9CDCFE',
            },
        },
        {
            'name': 'CSS property value',
            'scope': [
                'support.constant.property-value',
                'support.constant.font-name',
                'support.constant.media-type',
                'support.constant.media',
                'constant.other.color.rgb-value',
                'constant.other.rgb-value',
                'support.constant.color',
            ],
            'settings': {
                'foreground': '#CE9178',
            },
        },
        {
            'name': 'Regular expression groups',
            'scope': [
                'punctuation.definition.group.regexp',
                'punctuation.definition.group.assertion.regexp',
                'punctuation.definition.character-class.regexp',
                'punctuation.character.set.begin.regexp',
                'punctuation.character.set.end.regexp',
                'keyword.operator.negation.regexp',
                'support.other.parenthesis.regexp',
            ],
            'settings': {
                'foreground': '#CE9178',
            },
        },
        {
            'scope': [
                'constant.character.character-class.regexp',
                'constant.other.character-class.set.regexp',
                'constant.other.character-class.regexp',
                'constant.character.set.regexp',
            ],
            'settings': {
                'foreground': '#d16969',
            },
        },
        {
            'scope': [
                'keyword.operator.or.regexp',
                'keyword.control.anchor.regexp',
            ],
            'settings': {
                'foreground': '#DCDCAA',
            },
        },
        {
            'scope': 'keyword.operator.quantifier.regexp',
            'settings': {
                'foreground': '#d7ba7d',
            },
        },
        {
            'scope': 'constant.character',
            'settings': {
                'foreground': '#569cd6',
            },
        },
        {
            'scope': 'constant.character.escape',
            'settings': {
                'foreground': '#d7ba7d',
            },
        },
        {
            'scope': 'entity.name.label',
            'settings': {
                'foreground': '#C8C8C8',
            },
        },
        {
            'scope': 'entity.name.section.group-title',
            'settings': {
                'foreground': '#4EC9B0',
                'fontStyle': 'bold',
            },
        },
    ],
}



from __future__ import annotations

import collections
import contextlib
import os.path
from typing import Generator

from babi.user_data import xdg_data


class History:
    def __init__(self) -> None:
        self._orig_len: dict[str, int] = collections.defaultdict(int)
        self.data: dict[str, list[str]] = collections.defaultdict(list)
        self.prev: dict[str, str] = {}

    @contextlib.contextmanager
    def save(self) -> Generator[None]:
        history_dir = xdg_data('history')
        os.makedirs(history_dir, exist_ok=True)
        for filename in os.listdir(history_dir):
            history_filename = os.path.join(history_dir, filename)
            with open(history_filename, encoding='UTF-8') as f:
                self.data[filename] = f.read().splitlines()
                self._orig_len[filename] = len(self.data[filename])
        try:
            yield
        finally:
            for k, v in self.data.items():
                new_history = v[self._orig_len[k]:]
                if new_history:
                    history_filename = os.path.join(history_dir, k)
                    with open(history_filename, 'a+', encoding='UTF-8') as f:
                        f.write('\n'.join(new_history) + '\n')



from __future__ import annotations

import collections
import curses
import functools
from typing import TYPE_CHECKING

from babi.buf import Buf
from babi.color_manager import ColorManager
from babi.dim import Dim
from babi.highlight import Compiler
from babi.highlight import Grammar
from babi.highlight import Grammars
from babi.highlight import highlight_line
from babi.hl.interface import HL
from babi.hl.interface import HLs
from babi.horizontal_scrolling import scrolled_line
from babi.linting import Error
from babi.theme import Theme

if TYPE_CHECKING:
    from babi.main import Screen  # XXX: circular


@functools.lru_cache(maxsize=1)
def _compiler() -> Compiler:
    grammar = Grammar.make({
        'scopeName': 'source.demo',
        'fileTypes': ['demo'],
        'patterns': [
            {
                'match': r'^([0-9]+)(:)([0-9]+)(:) (\[[^\]]+\]) ([^ ]+)',
                'name': 'line.error',
                'captures': {
                    '1': {'name': 'constant.numeric'},
                    '2': {'name': 'punctuation.separator'},
                    '3': {'name': 'constant.numeric'},
                    '4': {'name': 'punctuation.separator'},
                    '5': {'name': 'strong support.type'},
                    '6': {'name': 'invalid'},
                },
            },
        ],
    })
    return Compiler(grammar, Grammars())


class LintErrors:
    include_edge = False

    def __init__(self, color_manager: ColorManager, theme: Theme) -> None:
        self._color_manager = color_manager
        self._theme = theme

        self.top = 0
        self.y = 0
        self.errors: tuple[Error, ...] = ()
        self._temporary_highlight = False

        self.regions: dict[int, HLs] = collections.defaultdict(tuple)

    def highlight_until(self, lines: Buf, idx: int) -> None:
        """our highlight regions are populated in other ways"""

    def _set_cb(self, lines: Buf, idx: int, victim: str) -> None:
        errors = tuple(
            error._replace(disabled=True) if error.line_idx == idx else error
            for error in self.errors
        )
        self.set_errors(errors)

    def _del_cb(self, lines: Buf, idx: int, victim: str) -> None:
        errors = tuple(
            error._replace(lineno=error.lineno - 1, disabled=True)
            if error.line_idx == idx else
            error._replace(lineno=error.lineno - 1)
            if error.line_idx > idx else
            error
            for error in self.errors
        )
        self.set_errors(errors)

    def _ins_cb(self, lines: Buf, idx: int) -> None:
        errors = tuple(
            error._replace(lineno=error.lineno + 1)
            if error.line_idx >= idx - 1 else error
            for error in self.errors
        )
        self.set_errors(errors)

    def register_callbacks(self, buf: Buf) -> None:
        buf.add_set_callback(self._set_cb)
        buf.add_del_callback(self._del_cb)
        buf.add_ins_callback(self._ins_cb)

    def set_errors(self, errors: tuple[Error, ...]) -> None:
        pair = self._color_manager.raw_color_pair(-1, curses.COLOR_RED)
        attr = curses.color_pair(pair)

        self.errors = errors
        self.regions.clear()
        self.regions.update({
            error.line_idx: (HL(x=0, end=1, attr=attr),)
            for error in errors
            if not error.disabled
        })

        if not self.errors:
            self.y = self.top = 0
        elif self.y >= len(self.errors):
            self.y = len(self.errors) - 1
            self.top = self.y - 1

    def clone(self, color_manager: ColorManager, theme: Theme) -> LintErrors:
        ret = type(self)(color_manager, theme)
        ret.set_errors(self.errors)
        return ret

    def draw(
            self,
            stdscr: curses._CursesWindow,
            dim: Dim,
            *,
            focused: bool = False,
    ) -> None:
        to_display = min(len(self.errors) - self.top, dim.height)
        for i in range(to_display):
            draw_y = i + dim.y
            l_y = self.top + i
            s = self.errors[l_y].render()
            rendered = scrolled_line(s, 0, dim.width)
            stdscr.insstr(draw_y, 0, rendered)

            if (self._temporary_highlight or focused) and self.y == l_y:
                attr = curses.A_REVERSE | curses.A_DIM | curses.color_pair(1)
                stdscr.chgat(draw_y, 0, dim.width, attr)
            elif self.errors[l_y].disabled:
                attr = curses.A_DIM | curses.color_pair(1)
                stdscr.chgat(draw_y, 0, dim.width, attr)
            else:
                compiler = _compiler()
                _, regions = highlight_line(
                    compiler, compiler.root_state, s, first_line=True,
                )

                # handle the scroll indicator
                if len(s) >= dim.width:
                    max_x = dim.width - 1
                else:
                    max_x = dim.width

                for r in regions:
                    style = self._theme.select(r.scope)
                    attr = style.attr(self._color_manager)
                    if r.start >= max_x:
                        break
                    stdscr.chgat(
                        draw_y,
                        r.start,
                        min(r.end, max_x) - r.start,
                        attr,
                    )

        for i in range(to_display, dim.height):
            stdscr.move(i + dim.y, 0)
            stdscr.clrtoeol()

        self._temporary_highlight = False

    def _up(self) -> None:
        self.y = max(self.y - 1, 0)
        if self.y < self.top:
            self.top -= 2

    def _down(self, dim: Dim) -> None:
        self.y = min(self.y + 1, len(self.errors) - 1)
        if self.top + dim.height <= self.y:
            self.top += 2

    def _set_file_position(self, screen: Screen, error: Error) -> None:
        screen.file.buf.assign_position(
            screen.layout.file,
            y=max(error.lineno - 1, 0),
            x=max(error.col_offset - 1, 0),
        )

    def _move_to_y(self, y: int, dim: Dim) -> None:
        while self.y > y:
            self._up()
        while self.y < y:
            self._down(dim)

    def previous_error(self, screen: Screen) -> None:
        if not self.errors:
            return

        pos = (screen.file.buf.y + 1, screen.file.buf.x + 1)

        y = -1
        for i, error in enumerate(self.errors):
            if error.disabled:
                continue
            elif error.pos >= pos:
                break
            else:
                y = i

        if y == -1:  # no previous errors
            self._temporary_highlight = pos == (self.errors[self.y].pos)
            return

        self._move_to_y(y, screen.layout.lint_errors)
        self._temporary_highlight = True
        self._set_file_position(screen, self.errors[self.y])

    def next_error(self, screen: Screen) -> None:
        if not self.errors:
            return

        pos = (screen.file.buf.y + 1, screen.file.buf.x + 1)

        y = -1
        for i, error in enumerate(reversed(self.errors)):
            if error.disabled:
                continue
            elif error.pos <= pos:
                break
            else:
                y = len(self.errors) - 1 - i

        if y == -1:  # no subsequent errors
            self._temporary_highlight = pos == (self.errors[self.y].pos)
            return

        self._move_to_y(y, screen.layout.lint_errors)
        self._temporary_highlight = True
        self._set_file_position(screen, self.errors[self.y])

    def focus(self, screen: Screen) -> None:
        if not self.errors:
            return

        while True:
            error = self.errors[self.y]
            if not error.disabled:
                self._set_file_position(screen, error)

            screen.draw()
            self.draw(screen.stdscr, screen.layout.lint_errors, focused=True)
            screen.file.move_cursor(screen.stdscr, screen.layout.file)

            ch = screen.get_char()
            if ch.keyname == b'KEY_RESIZE':
                screen.resize()
            elif ch.keyname in {b'^C', b'^X'}:
                self.set_errors(())
                screen.resize()
                return
            elif ch.keyname in {b'^[', b'M-t'}:
                return
            elif ch.keyname == b'^T':
                screen.lint()
                if not self.errors:
                    return
            elif ch.keyname == b'KEY_UP':
                self._up()
            elif ch.keyname == b'KEY_DOWN':
                self._down(screen.layout.lint_errors)



from __future__ import annotations

import functools
import os.path
import re
import subprocess
from typing import Tuple

from babi import linting

ErrorsByHook = Tuple[Tuple[str, Tuple[linting.Error, ...]], ...]

HOOK_ID_RE = re.compile('^- hook id: (.*)$')


def _parse_pre_commit(s: str) -> ErrorsByHook:
    ret = []
    current_hookid = ''
    current_lines: list[str] = []

    def _push_current_hook_id() -> None:
        nonlocal current_hookid

        if not current_hookid:
            return

        parsed = linting.parse_generic_output('\n'.join(current_lines))
        if parsed:
            ret.append((current_hookid, parsed))

    for line in s.splitlines():
        hook_id_match = HOOK_ID_RE.match(line)
        if hook_id_match:
            _push_current_hook_id()
            current_hookid = hook_id_match[1]
            current_lines.clear()
        else:
            current_lines.append(line)

    _push_current_hook_id()

    return tuple(ret)


class PreCommit:
    def __init__(self) -> None:
        self._root = functools.lru_cache(maxsize=None)(self._root_uncached)

    def _root_uncached(self, filename: str) -> str:
        return subprocess.check_output(
            (
                'git', '-C', os.path.dirname(os.path.abspath(filename)),
                'rev-parse', '--show-toplevel',
            ),
            text=True,
            stderr=subprocess.DEVNULL,
        ).rstrip()

    def command(self, filename: str, scope: str) -> tuple[str, ...] | None:
        try:
            root = self._root(filename)
        except subprocess.CalledProcessError:
            return None  # not in a git repo!

        # no pre-commit config!
        cfg = os.path.join(root, '.pre-commit-config.yaml')
        if not os.path.exists(cfg):
            return None

        return (
            'pre-commit', 'run',
            '--color=never',
            '--config', cfg,
            '--files', filename,
        )

    def parse(self, filename: str, output: str) -> tuple[linting.Error, ...]:
        root = self._root(filename)

        def _norm(path: str) -> str:
            return os.path.relpath(path, root)

        filename = _norm(filename)
        return tuple(
            error._replace(msg=f'[{hook_id}] {error.msg}')
            for hook_id, errors in _parse_pre_commit(output)
            for error in errors
            if _norm(os.path.join(root, error.filename)) == filename
        )



from __future__ import annotations

import curses

from babi.dim import Dim
from babi.prompt import PromptResult


class Status:
    def __init__(self) -> None:
        self._status = ''
        self._action_counter = -1

    def update(self, status: str) -> None:
        self._status = status
        self._action_counter = 25

    def clear(self) -> None:
        self._status = ''

    def draw(self, stdscr: curses._CursesWindow, dim: Dim) -> None:
        if dim.y > 0 or self._status:
            stdscr.insstr(dim.y, 0, ' ' * dim.width)
            if self._status:
                status = f' {self._status} '
                x = (dim.width - len(status)) // 2
                if x < 0:
                    x = 0
                    status = status.strip()
                stdscr.insstr(dim.y, x, status, curses.A_REVERSE)

    def tick(self, dim: Dim) -> None:
        # when the window is only 1-tall, hide the status quicker
        if dim.y > 0:
            self._action_counter -= 1
        else:
            self._action_counter -= 24
        if self._action_counter < 0:
            self.clear()

    def cancelled(self) -> PromptResult:
        self.update('cancelled')
        return PromptResult.CANCELLED
